# ===========================================================================
# Copyright 2013 University of Limerick
#
# This file is part of DREAM.
#
# DREAM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DREAM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with DREAM.  If not, see <http://www.gnu.org/licenses/>.
# ===========================================================================
'''
Created on 8 Nov 2012

@author: George
'''
'''
Models a machine that can also have failures
'''

# from SimPy.Simulation import Process, Resource, SimEvent
# from SimPy.Simulation import activate, passivate, waituntil, now, hold, request, release, waitevent
import simpy

from Failure import Failure
from CoreObject import CoreObject

from OperatorRouter import Router
from SkilledOperatorRouter import SkilledRouter

from OperatedPoolBroker import Broker
from OperatorPool import OperatorPool

from RandomNumberGenerator import RandomNumberGenerator

# ===========================================================================
# the Machine object
# ===========================================================================
class Machine(CoreObject):
    family='Server'
    
    def getProcessingTime(self,processingTime):
        '''returns the processingTime dictionary updated'''
        if not processingTime:
            processingTime = { 'distributionType': 'Fixed',
                               'mean': 1, }
        if processingTime['distributionType'] == 'Normal' and\
                processingTime.get('max', None) is None:
            processingTime['max'] = float(processingTime['mean']) + 5 * float(processingTime['stdev'])
        return processingTime
        
    def getSetupTime(self,setupTime):
        '''returns the setupTime dictionary updated'''
        if not setupTime:
            setupTime = { 'distributionType': 'Fixed',
                          'mean': 1, }
        if setupTime['distributionType'] == 'Normal' and\
                setupTime.get('max', None) is None:
            setupTime['max'] = float(setupTime['mean']) + 5 * float(setupTime['stdev'])
        return setupTime
          
    def getLoadTime(self,loadTime):
        '''returns the loadTime dictionary updated'''
        if not loadTime:
            loadTime = { 'distributionType': 'Fixed',
                         'mean': 1, }
        if loadTime['distributionType'] == 'Normal' and\
                loadTime.get('max', None) is None:
            loadTime['max'] = float(loadTime['mean']) + 5 * float(loadTime['stdev'])
        return loadTime
    
    # =======================================================================
    # initialise the id the capacity, of the resource and the distribution
    # =======================================================================
    def __init__(self, id, name, capacity=1, processingTime=None,
                  repairman='None',\
                  operatorPool='None',operationType='None',\
                  setupTime=None, loadTime=None,
                  preemption={},
                  canDeliverOnInterruption=False, **kw):
        self.type="Machine"                         #String that shows the type of object
        CoreObject.__init__(self, id, name)
        from Globals import G
        
        processingTime=self.getProcessingTime(processingTime)
        
        setupTime=self.getSetupTime(setupTime)
        
        loadTime=self.getLoadTime(loadTime)
        
        #     holds the capacity of the machine 
        self.capacity=capacity
        #     sets the repairman resource of the Machine
        self.repairman=repairman
        #     Sets the attributes of the processing (and failure) time(s)
        self.rng=RandomNumberGenerator(self, **processingTime)
        # check whether the operators are provided with a skills set
        # check whether the operators are provided with a skills set
        self.dedicatedOperator=self.checkForDedicatedOperators()
        if len(G.OperatorPoolsList)>0:
            for operatorPool in G.OperatorPoolsList:                    # find the operatorPool assigned to the machine
                if(self.id in operatorPool.coreObjectIds):                   # and add it to the machine's operatorPool
                    machineOperatorPoolList=operatorPool                # there must only one operator pool assigned to the machine,
                                                                        # otherwise only one of them will be taken into account
                else:
                    machineOperatorPoolList=[]                          # if there is no operatorPool assigned to the machine
        else:                                                           # then machineOperatorPoolList/operatorPool is a list
            machineOperatorPoolList=[]                                  # if there are no operatorsPool created then the 
                                                                        # then machineOperatorPoolList/operatorPool is a list
        if (type(machineOperatorPoolList) is list):                 # if the machineOperatorPoolList is a list
                                                                    # find the operators assigned to it and add them to the list
            for operator in G.OperatorsList:                        # check which operator in the G.OperatorsList
                if(self.id in operator.coreObjectIds):                   # (if any) is assigned to operate
                    machineOperatorPoolList.append(operator)        # the machine with ID equal to id
        
        self.operatorPool=machineOperatorPoolList
        
        self.dedicatedOperator=self.checkForDedicatedOperators()
        # create an operatorPool if needed
        self.createOperatorPool(self.operatorPool)
        # holds the Operator currently processing the Machine
        self.currentOperator=None
        # define if load/setup/removal/processing are performed by the operator 
        self.operationType=operationType
        # boolean to check whether the machine is being operated
        self.toBeOperated = False
        # define the load times
        self.loadRng = RandomNumberGenerator(self, **loadTime)
        # variable that informs on the need for setup
        self.setUp=True
        # define the setup times
        self.stpRng = RandomNumberGenerator(self, **setupTime)
        # examine if there are multiple operation types performed by the operator
        #     there can be Setup/Processing operationType
        #     or the combination of both (MT-Load-Setup-Processing) 
        self.multOperationTypeList=[]   
        if self.operationType.startswith("MT"):
            OTlist = operationType.split('-')
            self.operationType=OTlist.pop(0)
            self.multOperationTypeList = OTlist
        else:
            self.multOperationTypeList.append(self.operationType)
 
        # flags used for preemption purposes
        self.isPreemptive=False
        self.resetOnPreemption=False
        if len(preemption)>0:
            self.isPreemptive=bool(int(preemption.get('isPreemptive') or 0))
            self.resetOnPreemption=bool(int(preemption.get('resetOnPreemption', 0)))
        # flag notifying that there is operator assigned to the actievObject
        self.assignedOperator=True
        # flag notifying the the station can deliver entities that ended their processing while interrupted
        self.canDeliverOnInterruption=canDeliverOnInterruption
        self.repairman='None'
        for repairman in G.RepairmanList:                   # check which repairman in the G.RepairmanList
            if(self.id in repairman.coreObjectIds):              # (if any) is assigned to repair 
                self.repairman=repairman                                 # the machine with ID equal to id
        G.MachineList.append(self)                             # add machine to global MachineList
        if self.operatorPool!="None":
            G.OperatedMachineList.append(self)                 # add the machine to the operatedMachines List

   
    # =======================================================================
    # initialize the machine
    # =======================================================================        
    def initialize(self):
        # using the Process __init__ and not the CoreObject __init__
        CoreObject.initialize(self)
        
        # initialize the internal Queue (type Resource) of the Machine 
        self.Res=simpy.Resource(self.env, capacity=1)
        # initiate the Broker and the router
        self.createBroker()
        self.createRouter()
        # initialize the operator pool if any
        self.initializeOperatorPool()
        # initiate the Broker responsible to control the request/release
        self.initializeBroker()
        # initialise the router if not initialized already
        self.initializeRouter()
        
        # TODO: check whether the requestingEntity variable can be used in OperatorPreemptive
        self.requestingEntity=None
        # variables used for interruptions
        self.tinM=0
        self.timeLastProcessingStarted=0
        self.interruption=False
        self.breakTime=0
        # flag notifying that there is operator assigned to the actievObject
        self.assignedOperator=True
        
        self.brokerIsSet=self.env.event()
        # this event is generated every time an operator is requested by machine for Load operation type.
        #     if the machine has not succeeded in getting an entity due to the resource absence 
        #     and waits for the next event to get the entity, 
        #     then it must be signalled that the operator is now available
        self.loadOperatorAvailable=self.env.event()
        # signal used for preemption
        self.preemptQueue=self.env.event()
        # signal used for informing objectInterruption objects that the current entity processed has finished processnig
        self.endedLastProcessing=self.env.event()
                
    #===========================================================================
    # create an operatorPool if needed
    #===========================================================================
    def createOperatorPool(self,operatorPool):
        ''' sets the operator resource of the Machine
             check if the operatorPool is a List or a OperatorPool type Object
             if it is a list then initiate a OperatorPool type object containing
             the list of operators provided
            if the  list is empty create operator pool with empty list
        '''
        from Globals import G
        # XXX operatorPool is not None ?
        # if a list of operators is provided as argument
        if (type(operatorPool) is list) and len(operatorPool)>0:
            id = self.id+'_OP'
            name=self.objName+'_operatorPool'
            self.operatorPool=OperatorPool(id, name, operatorsList=operatorPool)
            G.OperatorPoolsList.append(self.operatorPool)
        # if an operatorPool object is connected to the machine
        elif(type(operatorPool) is OperatorPool):
            self.operatorPool=operatorPool
        # otherwise
        else:
            # if there are operators with skillList
            if self.dedicatedOperator:
                id = self.id+'_OP'
                name=self.objName+'_operatorPool'
                self.operatorPool=OperatorPool(id,name,operatorsList=[])
                G.OperatorPoolsList.append(self.operatorPool)
            # otherwise create no operatorPool
            else:
                self.operatorPool='None'
        # update the operatorPool coreObjects list
        if self.operatorPool!='None':
             self.operatorPool.coreObjectIds.append(self.id)
             self.operatorPool.coreObjects.append(self)
    
    #===========================================================================
    # create broker if needed
    #===========================================================================
    def createBroker(self):
        # initiate the Broker and the router
        if (self.operatorPool!='None'):
            self.broker=Broker(self)
    
    #===========================================================================
    # create router if needed
    #===========================================================================
    def createRouter(self):
        # initiate the Broker and the router
        if (self.operatorPool!='None'):
            from Globals import G
            # if there is no router
            if not G.Router:
                # TODO if the dedicatedOperator flag is raised then create a SkilledRouter (temp)
                if self.dedicatedOperator:
                    self.router=SkilledRouter()
                else:
                    self.router=Router()
                G.Router=self.router
            # otherwise set the already existing router as the machines Router
            else:
                self.router=G.Router
    #===========================================================================
    # initialise broker if needed
    #===========================================================================
    def initializeOperatorPool(self):
        # initialise the operator pool if any
        if (self.operatorPool!="None"):
            self.operatorPool.initialize()
            # if the flag dedicatedOperator is true then reset/empty the operators list of the pool
            if self.dedicatedOperator:
                self.operatorPool.operators=[]
            # otherwise update the coreObjectIds/coreObjects list of the operators
            else:
                for operator in self.operatorPool.operators:
                    operator.coreObjectIds.append(self.id)
                    operator.coreObjects.append(self)
    
    #===========================================================================
    # initialise broker if needed
    #===========================================================================
    def initializeBroker(self):
        # initiate the Broker responsible to control the request/release
        if (self.operatorPool!="None"):
            self.broker.initialize()
            self.env.process(self.broker.run())
                
    #===========================================================================
    # initialise router if needed
    #===========================================================================
    def initializeRouter(self):
        if (self.operatorPool!="None"):
            # initialise the router only once
            if not self.router.isInitialized:
                self.router.initialize()
            if not self.router.isActivated:
                self.env.process(self.router.run())
                self.router.isActivated=True
                
    # =======================================================================
    # the main process of the machine
    # =======================================================================
    def run(self):
        # check if there is WIP and signal receiver
        self.initialSignalReceiver()
        # execute all through simulation time
        while 1:
            # waitEvent isRequested /interruptionEnd/loadOperatorAvailable
            while 1:
                self.printTrace(self.id, waitEvent='')
                receivedEvent = yield self.env.any_of([self.isRequested, self.interruptionEnd, 
                                                       self.loadOperatorAvailable, self.initialWIP])
                self.printTrace(self.id, received='')
                # if the machine can accept an entity and one predecessor requests it continue with receiving the entity
                if self.isRequested in receivedEvent:
                    self.printTrace(self.id, isRequested=self.isRequested.value.id)
                    assert self.isRequested.value==self.giver, 'the giver is not the requestingObject'
                    assert self.giver.receiver==self, 'the receiver of the signalling object in not the station'
                    # reset the signalparam of the isRequested event
                    self.isRequested=self.env.event()
                    break
                # if an interruption caused the control to be taken by the machine or
                # if an operator was rendered available while it was needed by the machine to proceed with getEntity
                if self.interruptionEnd in receivedEvent or self.loadOperatorAvailable in receivedEvent:
                    if self.interruptionEnd in receivedEvent:
                        assert self.interruptionEnd.value==self.env.now, 'interruptionEnd received later than created'
                        self.printTrace(self.id, interruptionEnd=str(self.interruptionEnd.value))
                        self.interruptionEnd=self.env.event()
                    if self.loadOperatorAvailable in receivedEvent:
                        assert self.loadOperatorAvailable.value==self.env.now,'loadOperatorAvailable received later than created'
                        self.printTrace(self.id,loadOperatorAvailable=str(self.loadOperatorAvailable.value))
                        self.loadOperatorAvailable=self.env.event()
                    # try to signal the Giver, otherwise wait until it is requested
                    if self.signalGiver():
                        break
                if self.initialWIP in receivedEvent:
                    assert self.initialWIP.value==self.env, 'initial wip was not sent by the Environment'
                    self.initialWIP=self.env.event()
                    self.isProcessingInitialWIP=True
                    break
                    
            # TODO: maybe here have to assigneExit of the giver and add self to operator activeCallers list
                
            # here or in the getEntity (apart from the loadTimeCurrentEntity)
            # in case they are placed inside the getEntity then the initialize of
            # the corresponding variables should be moved to the initialize() of the CoreObject
            self.operatorWaitTimeCurrentEntity = 0
            self.loadOperatorWaitTimeCurrentEntity = 0
            self.loadTimeCurrentEntity = 0
            self.setupTimeCurrentEntity = 0
            
    # ======= request a resource
            if(self.operatorPool!="None") and any(type=='Load' for type in self.multOperationTypeList):
                # when it's ready to accept (canAcceptAndIsRequested) then inform the broker
                # machines waits to be operated (waits for the operator)
                self.requestOperator()
                self.timeWaitForLoadOperatorStarted = self.env.now
                # wait until the Broker has waited times equal to loadTime (if any)
                yield self.brokerIsSet
                self.brokerIsSet=self.env.event()
                self.timeWaitForLoadOperatorEnded = self.env.now
                self.loadOperatorWaitTimeCurrentEntity += self.timeWaitForLoadOperatorEnded-self.timeWaitForLoadOperatorStarted
                self.totalTimeWaitingForLoadOperator += self.loadOperatorWaitTimeCurrentEntity 
                 
    # ======= Load the machine if the Load is defined as one of the Operators' operation types
            if any(type=="Load" for type in self.multOperationTypeList) and self.isOperated():
                self.timeLoadStarted = self.env.now
                yield self.env.timeout(self.calculateLoadTime())
                # TODO: if self.interrupted(): There is the issue of failure during the Loading
                self.timeLoadEnded = self.env.now
                self.loadTimeCurrentEntity = self.timeLoadEnded-self.timeLoadStarted 
                self.totalLoadTime += self.loadTimeCurrentEntity
                 
    # ======= release a resource if the only operation type is Load
            if (self.operatorPool!="None")\
                 and any(type=="Load" for type in self.multOperationTypeList)\
                 and not (any(type=="Processing" for type in self.multOperationTypeList)\
                          or any(type=="Setup" for type in self.multOperationTypeList))\
                 and self.isOperated():
                # after getting the entity release the operator
                # machine has to release the operator
                self.releaseOperator()
                # wait until the Broker has finished processing
                yield self.brokerIsSet
                self.brokerIsSet=self.env.event()
             
            # TODO: reset the requestinEntity before receiving the currentEntity
            self.requestingEntity=None
            # get the entity 
                    # TODO: if there was loading time then we must solve the problem of getting an entity
                    # from an unidentified giver or not getting an entity at all as the giver 
                    # may fall in failure mode (assignExit()?)
            if not self.isProcessingInitialWIP:     # if we are in the state of having initial wip no need to take an Entity
                self.currentEntity=self.getEntity()
            # TODO: the Machine receive the entity  after the operator is available
            #     the canAcceptAndIsRequested method checks only in case of Load type of operation 
            
    # ======= request a resource if it is not already assigned an Operator
            if(self.operatorPool!="None")\
                 and (any(type=="Processing" for type in self.multOperationTypeList)\
                      or any(type=="Setup" for type in self.multOperationTypeList))\
                 and not self.isOperated():
                # when it's ready to accept (canAcceptAndIsRequested) then inform the broker
                # machines waits to be operated (waits for the operator)
                self.requestOperator()
                self.timeWaitForOperatorStarted = self.env.now
                # wait until the Broker has waited times equal to loadTime (if any)
                yield self.brokerIsSet
                self.brokerIsSet=self.env.event()
                self.timeWaitForOperatorEnded = self.env.now
                self.operatorWaitTimeCurrentEntity += self.timeWaitForOperatorEnded-self.timeWaitForOperatorStarted
            
            # variables dedicated to hold the processing times, the time when the Entity entered, 
            # and the processing time left 
            self.totalProcessingTimeInCurrentEntity=self.calculateProcessingTime()                # get the processing time, tinMStarts holds the processing time of the machine 
            self.tinM=self.totalProcessingTimeInCurrentEntity                                          # timer to hold the processing time left
            
    # ======= setup the machine if the Setup is defined as one of the Operators' operation types
            # in plantSim the setup is performed when the machine has to process a new type of Entity and only once
            if any(type=="Setup" for type in self.multOperationTypeList) and self.isOperated():
                self.timeSetupStarted = self.env.now
                yield self.env.timeout(self.calculateSetupTime())
                # TODO: if self.interrupted(): There is the issue of failure during the setup
                self.timeSetupEnded = self.env.now
                self.setupTimeCurrentEntity = self.timeSetupEnded-self.timeSetupStarted
                self.totalSetupTime += self.setupTimeCurrentEntity
                 
            # setup is performed only when setup is set in Machines multOperationTypeList
            # TODO: This must be also performed when no such operation is defined for the operator
            #     but setupTime is given for the entity to be processed
#             try:
#                 if self.setupTime and not any(type=="Setup" for type in self.multOperationTypeList):
#                     self.timeSetupStarted = self.env.now
#                     yield hold,self,self.calculateSetupTime()
#                     # TODO: if self.interrupted(): There is the issue of failure during the setup
#                     self.timeSetupEnded = self.env.now
#                     self.setupTimeCurrentEntity = self.timeSetupEnded-self.timeSetupStarted
#                     self.totalSetupTime += self.setupTimeCurrentEntity
#             except:
#                 pass
             
    # ======= release a resource if the only operation type is Setup
            if (self.operatorPool!="None")\
                and self.isOperated()\
                and (any(type=="Setup" for type in self.multOperationTypeList)\
                     or any(type=="Load" for type in self.multOperationTypeList))\
                and not any(type=="Processing" for type in self.multOperationTypeList):
                # after getting the entity release the operator
                # machine has to release the operator
                self.releaseOperator()
                # wait until the Broker has finished processing
                yield self.brokerIsSet
                self.brokerIsSet=self.env.event()
                                                                                 
            # variables used to flag any interruptions and the end of the processing
            self.interruption=False
            processingNotFinished=True
            # timers to follow up the failure time of the machine while on current Entity
            self.downTimeInCurrentEntity=0                          #holds the total time that the 
                                                                    #object was down while holding current entity
            
            # if there is a failure that depends on the working time of the Machine
            # send it the victimStartsProcess signal                                            
            for oi in self.objectInterruptions:
                if oi.type=='Failure':
                    if oi.deteriorationType=='working':
                        oi.victimStartsProcess.succeed(self.env.now)
                                                 
            # this loop is repeated until the processing time is expired with no failure
            # check when the processingEndedFlag switched to false              
            while processingNotFinished:
                # timeLastProcessingStarted : dummy variable to keep track of the time that the processing starts after 
                #           every interruption                        
                self.timeLastProcessingStarted=self.env.now
                # processing starts, update the flag
                self.isProcessing=True
                # wait for the processing time left tinM, if no interruption occurs then change the processingEndedFlag and exit loop,
                #     else (if interrupted()) set interruption flag to true (only if tinM==0), 
                #     and recalculate the processing time left tinM, passivate while waiting for repair.
                # if a preemption has occurred then react accordingly (proceed with getting the critical entity)
                # receivedEvent = yield self.env.timeout(self.tinM) | self.interruptionStart | self.preemptQueue
                receivedEvent = yield self.env.any_of([self.interruptionStart, self.env.timeout(self.tinM) , self.preemptQueue])
                if self.interruptionStart in receivedEvent:                              # if a failure occurs while processing the machine is interrupted.
                    assert self.interruptionStart.value==self.env.now, 'the interruption has not been processed on the time of activation'
                    self.interruptionStart=self.env.event()
                    self.interruptionActions()                      # execute interruption actions
                    
    # =============== release the operator if there is interruption 
                    if (self.operatorPool!="None")\
                        and self.isOperated()\
                        and any(type=="Processing" for type in self.multOperationTypeList):
                        self.releaseOperator()
                        yield self.brokerIsSet
                        self.brokerIsSet=self.env.event()
    
                    # loop until we reach at a state that there is no interruption
                    while 1:
                        yield self.interruptionEnd         # interruptionEnd to be triggered by ObjectInterruption
                        assert self.env.now==self.interruptionEnd.value, 'the victim of the failure is not the object that received it'
                        self.interruptionEnd=self.env.event()
                        if self.Up and self.onShift:
                            break
                    self.postInterruptionActions()
                    
                
    # =============== request a resource after the repair
                    if (self.operatorPool!="None")\
                        and any(type=="Processing" for type in self.multOperationTypeList)\
                        and not self.interruption:
                        self.timeWaitForOperatorStarted = self.env.now
                        self.requestOperator()
                        yield self.brokerIsSet
                        self.brokerIsSet=self.env.event()
                        self.timeWaitForOperatorEnded = self.env.now 
                        self.operatorWaitTimeCurrentEntity += self.timeWaitForOperatorEnded-self.timeWaitForOperatorStarted
                # if the station is reactivated by the preempt method
                elif(self.shouldPreempt):
                    if (self.preemptQueue in receivedEvent):
                        assert self.preemptQueue.value==self.env.now, 'the preemption must be performed on the time of request'
                        self.preemptQueue=self.env.event()
                    self.interruptionActions()                      # execute interruption actions
                         
    # =============== release the operator if there is interruption 
                    if (self.operatorPool!="None")\
                        and self.isOperated()\
                        and any(type=="Processing" for type in self.multOperationTypeList):
                        self.releaseOperator()
                        yield self.brokerIsSet
                        self.brokerIsSet=self.env.event()
                     
                    self.postInterruptionActions()
                    break
                
                # if no interruption occurred the processing in M1 is ended 
                else:
                    processingNotFinished=False
            
            # carry on actions that have to take place when an Entity ends its processing
            self.endProcessingActions()
    # =============== release resource after the end of processing
            if (self.operatorPool!='None')\
                and self.isOperated()\
                and any(type=="Processing" for type in self.multOperationTypeList)\
                and not self.interruption: 
                self.releaseOperator()
                yield self.brokerIsSet
                self.brokerIsSet=self.env.event()
            # signal the receiver that the activeObject has something to dispose of
            if not self.signalReceiver():
            # if there was no available receiver, get into blocking control
                while 1:
                    self.timeLastBlockageStarted=self.env.now       # blockage is starting
                    # wait the event canDispose, this means that the station can deliver the item to successor
                    self.printTrace(self.id, waitEvent='(canDispose or interruption start)')
                    receivedEvent=yield self.env.any_of([self.canDispose , self.interruptionStart])
                    # if there was interruption
                    # TODO not good implementation
                    if self.interruptionStart in receivedEvent:
                        assert self.interruptionStart.value==self.env.now, 'the interruption has not been processed on the time of activation'
                        self.interruptionStart=self.env.event()
                    # wait for the end of the interruption
                        self.interruptionActions()                          # execute interruption actions
                        # loop until we reach at a state that there is no interruption
                        while 1:
                            yield self.interruptionEnd         # interruptionEnd to be triggered by ObjectInterruption
                            assert self.env.now==self.interruptionEnd.value, 'the victim of the failure is not the object that received it'
                            self.interruptionEnd=self.env.event()
                            if self.Up and self.onShift:
                                break
                        self.postInterruptionActions()
                        if self.signalReceiver():
                            self.timeLastBlockageStarted=self.env.now
                            break
                        else:
                            continue
                    if self.canDispose in receivedEvent:
                        if self.canDispose.value!=self.env.now:
                            self.canDispose=self.env.event()
                            continue
                        assert self.canDispose.value==self.env.now,'canDispose signal is late'
                        self.canDispose=self.env.event()
                        # try to signal a receiver, if successful then proceed to get an other entity
                        if self.signalReceiver():
                            break
                    # TODO: router most probably should signal givers and not receivers in order to avoid this hold,self,0
                    #       As the receiver (e.g.) a machine that follows the machine receives an loadOperatorAvailable event,
                    #       signals the preceding station (e.g. self.machine) and immediately after that gets the entity.
                    #       the preceding machine gets the canDispose signal which is actually useless, is emptied by the following station
                    #       and then cannot exit an infinite loop.
                    # notify that the station waits the entity to be removed
                    self.waitEntityRemoval=True
                    self.printTrace(self.id, waitEvent='(entityRemoved)')
                    yield self.entityRemoved
                    self.printTrace(self.id, entityRemoved=self.entityRemoved.value)
                    assert self.entityRemoved.value==self.env.now,'entityRemoved event activated earlier than received'
                    self.waitEntityRemoval=False
                    self.entityRemoved=self.env.event()
                    # if while waiting (for a canDispose event) became free as the machines that follows emptied it, then proceed
                    if not self.haveToDispose():
                        break
    
    # =======================================================================
    # actions to be carried out when the processing of an Entity ends
    # =======================================================================    
    def endProcessingActions(self):
        # set isProcessing to False
        self.isProcessing=False
        # add working time
        self.totalWorkingTime+=self.env.now-self.timeLastProcessingStarted

        # blocking starts
        self.isBlocked=True
        self.timeLastBlockageStarted=self.env.now
        
        activeObjectQueue=self.Res.users
        activeEntity=activeObjectQueue[0]
        self.printTrace(self.getActiveObjectQueue()[0].name, processEnd=self.objName)
        # reset the variables used to handle the interruptions timing 
        # self.timeLastProcessingStarted=0
        self.breakTime=0
        # output to trace that the processing in the Machine self.objName ended 
        try:
            self.outputTrace(activeObjectQueue[0].name,"ended processing in "+self.objName)
        except IndexError:
            pass
        # TODO: Not only Machines require time to process entities
        #     entities such as batchReassembly/Decomposition require time to process entities
        if activeEntity.family=='Entity':
            successorsAreMachines=True
            from Globals import G
            for object in self.next:
                if not object in G.MachineList:
                    successorsAreMachines=False
                    break
            if not successorsAreMachines:
                activeObjectQueue[0].hot = False
        from Globals import G
        if G.Router:
            # the just processed entity is added to the list of entities 
            # pending for the next processing
            G.pendingEntities.append(activeObjectQueue[0])
        # set the variable that flags an Entity is ready to be disposed 
        self.waitToDispose=True

        # update the variables keeping track of Entity related attributes of the machine    
        self.timeLastEntityEnded=self.env.now                          # this holds the time that the last entity ended processing in Machine 
        self.nameLastEntityEnded=self.currentEntity.name        # this holds the name of the last entity that ended processing in Machine
        self.completedJobs+=1                                   # Machine completed one more Job
        # reseting flags
        self.shouldPreempt=False
        self.isProcessingInitialWIP=False

        # if there is a failure that depends on the working time of the Machine
        # send it the victimEndsProcess signal                                  
        for oi in self.objectInterruptions:
            if oi.type=='Failure':
                if oi.deteriorationType=='working':
                    oi.victimEndsProcess.succeed(self.env.now)

        # in case Machine just performed the last work before the scheduled maintenance signal the corresponding object
        if self.isWorkingOnTheLast:
            # for the scheduled Object interruptions
            # XXX add the SkilledOperatorRouter to this list and perform the signalling only once
            for interruption in (G.ObjectInterruptionList):
                # if the objectInterruption is waiting for a a signal
                if interruption.victim==self and interruption.waitingSignal:
                    # signal it and reset the flags
                    self.endedLastProcessing.succeed(self.env.now)
                    interruption.waitinSignal=False
                    self.isWorkingOnTheLast=False
            # set timeLastShiftEnded attribute so that if it is overtime working it is not counted as off-shift time
            if self.interruptedBy=='ShiftScheduler':
                self.timeLastShiftEnded=self.env.now
    
    # =======================================================================
    # actions to be carried out when the processing of an Entity ends
    # =======================================================================    
    def interruptionActions(self):
        # if object was processing add the working time
        # only if object is not preempting though
        # in case of preemption endProcessingActions will be called
        if self.isProcessing and not self.shouldPreempt:
            self.totalWorkingTime+=self.env.now-self.timeLastProcessingStarted
        # if object was blocked add the working time
        if self.isBlocked:
            self.addBlockage()
            
        # set isProcessing to False          
        self.isProcessing=False
        # set isBlocked to False          
        self.isBlocked=False
        activeObjectQueue=self.Res.users
        activeEntity=activeObjectQueue[0]
        self.printTrace(activeEntity.name, interrupted=self.objName)
        # if the interrupt occurred while processing an entity
        if not self.waitToDispose:
            # output to trace that the Machine (self.objName) got interrupted           
            try:                                                       
                self.outputTrace(activeObjectQueue[0].name, "Interrupted at "+self.objName)
            except IndexError:
                pass
            # recalculate the processing time left tinM
            self.tinM=self.tinM-(self.env.now-self.timeLastProcessingStarted)
            if(self.tinM==0):       # sometimes the failure may happen exactly at the time that the processing would finish
                                    # this may produce disagreement with the simul8 because in both SimPy and Simul8
                                    # it seems to be random which happens 1st
                                    # this should not appear often to stochastic models though where times are random
                self.interruption=True
        # start counting the down time at breatTime dummy variable
        self.breakTime=self.env.now        # dummy variable that the interruption happened
    
    # =======================================================================
    # actions to be carried out when the processing of an Entity ends
    # =======================================================================    
    def postInterruptionActions(self):
        activeObjectQueue=self.Res.users
        activeEntity=activeObjectQueue[0]
        # if the machine returns from an failure while processing an entity
        if not self.waitToDispose:
            # use the timers to count the time that Machine is down and related 
            self.downTimeProcessingCurrentEntity+=self.env.now-self.breakTime      # count the time that Machine is down while processing this Entity
            self.downTimeInCurrentEntity+=self.env.now-self.breakTime              # count the time that Machine is down while on currentEntity
            self.timeLastFailureEnded=self.env.now                                 # set the timeLastFailureEnded
            self.failureTimeInCurrentEntity+=self.env.now-self.breakTime           # dummy variable keeping track of the failure time 
            # output to trace that the Machine self.objName was passivated for the current failure time
            self.outputTrace(activeObjectQueue[0].name, "passivated in "+self.objName+" for "+str(self.env.now-self.breakTime))
        # when a machine returns from failure while trying to deliver an entity
        else:
            # count the failure while on current entity time with failureTime variable
            self.failureTimeInCurrentEntity+=self.env.now-self.breakTime
            # calculate the time the Machine was down while trying to dispose the current Entity,
            # and the total down time while on current Entity
            self.downTimeInTryingToReleaseCurrentEntity+=self.env.now-self.breakTime
            self.downTimeInCurrentEntity+=self.env.now-self.breakTime        # already updated from failures during processing
            # update the timeLastFailureEnded
            self.timeLastFailureEnded=self.env.now
            # reset the variable holding the time the failure happened
            self.breakTime=0
                            
    # =======================================================================
    # checks if the machine is Up
    # =======================================================================
    def checkIfMachineIsUp(self):
        return self.Up 

    # =======================================================================
    # checks if the Machine can accept an entity
    # it checks also who called it and returns TRUE only to the predecessor 
    # that will give the entity.
    # =======================================================================  
    def canAccept(self, callerObject=None):
        if self.isLocked:
            return False
        activeObjectQueue=self.Res.users
        # if we have only one predecessor just check if there is a place and the machine is up
        # this is done to achieve better (cpu) processing time 
        # then we can also use it as a filter for a yield method
        if(callerObject==None):
            if (self.operatorPool!='None' and (any(type=='Load' for type in self.multOperationTypeList)\
                                                    or any(type=='Setup' for type in self.multOperationTypeList))):
                return self.operatorPool.checkIfResourceIsAvailable()\
                        and self.checkIfMachineIsUp()\
                        and len(activeObjectQueue)<self.capacity\
                        and not self.entryIsAssignedTo()
            else:
                return self.checkIfMachineIsUp()\
                        and len(activeObjectQueue)<self.capacity\
                        and not self.entryIsAssignedTo()
        thecaller=callerObject
        # return True ONLY if the length of the activeOjbectQue is smaller than
        # the object capacity, and the callerObject is not None but the giverObject
        if (self.operatorPool!='None' and (any(type=='Load' for type in self.multOperationTypeList)\
                                                or any(type=='Setup' for type in self.multOperationTypeList))):
            return self.operatorPool.checkIfResourceIsAvailable()\
                and self.checkIfMachineIsUp()\
                and len(activeObjectQueue)<self.capacity\
                and not self.entryIsAssignedTo()
        else:
            # the operator doesn't have to be present for the loading of the machine as the load operation
            # is not assigned to operators
            return self.checkIfMachineIsUp()\
                and len(activeObjectQueue)<self.capacity\
                and (thecaller in self.previous)\
                and not self.entryIsAssignedTo()
    
    # =======================================================================
    # checks if the Machine can accept an entity and there is an entity in 
    # some possible giver waiting for it
    # also updates the giver to the one that is to be taken
    # =======================================================================
    def canAcceptAndIsRequested(self,callerObject=None):
        activeObjectQueue=self.Res.users
        giverObject=callerObject
        assert giverObject, 'there must be a caller for canAcceptAndIsRequested'
        # check if there is a place, the machine is up and the predecessor has an entity to dispose. if the machine has to compete 
        # for an Operator that loads the entities onto it check if the predecessor if blocked by an other Machine. if not then the machine 
        # has to block the predecessor giverObject to avoid conflicts with other competing machines
        if (self.operatorPool!='None' and (any(type=='Load' for type in self.multOperationTypeList))):
            if giverObject.haveToDispose(self):
                if self.checkOperator()\
                    and self.checkIfActive() and len(activeObjectQueue)<self.capacity:
                    if not giverObject.exitIsAssignedTo():
                        giverObject.assignExitTo(self)
                    elif giverObject.exitIsAssignedTo()!=self:
                        return False
                    return True
            else:
                return False
        else:
            # the operator performs no load and the entity is received by the machine while there is 
            # no need for operators presence. The operator needs to be present only where the load Type 
            # operation is assigned
            if self.checkIfActive() and len(activeObjectQueue)<self.capacity\
                    and giverObject.haveToDispose(self):
                if not giverObject.exitIsAssignedTo():
                    giverObject.assignExitTo(self)
                elif giverObject.exitIsAssignedTo()!=self:
                    return False
                return True
    
    #===========================================================================
    # return whether Load or setup Requested
    #===========================================================================
    def isLoadRequested(self):
        return any(type=='Load' or type=='Setup' for type in self.multOperationTypeList)
    
    # =======================================================================
    # to be called by canAcceptAndIsRequested and check for the operator
    # =======================================================================    
    def checkOperator(self,callerObject=None):
        mayProceed=False
        if self.operatorPool.operators:
            # flag notifying that there is operator assigned to the actievObject
            self.assignedOperator=False
            # the operators operating the station
            operators=self.operatorPool.operators
            if self.operatorPool.checkIfResourceIsAvailable():
                for operator in [x for x in operators if x.checkIfResourceIsAvailable()]:
                    # if there are operators available not assigned to the station then the station may proceed signalling the Router
                    if not operator.isAssignedTo():
                        mayProceed=True
                    # if there are operators assigned to the station then proceed without invoking the Router
                    elif operator.isAssignedTo()==self:
                        self.assignedOperator=True
                        break
            return mayProceed or self.assignedOperator
        else:
            return True
    
    # =======================================================================
    # get an entity from the giver
    # =======================================================================
    def getEntity(self):
        activeEntity=CoreObject.getEntity(self)          # run the default method   
        # after the machine receives an entity, it must be removed from the pendingEntities list
        from Globals import G
        if G.Router:
            if activeEntity in G.pendingEntities:
                G.pendingEntities.remove(activeEntity)
        return activeEntity
  
    # =======================================================================
    # removes an entity from the Machine
    # =======================================================================
    def removeEntity(self, entity=None):
        activeEntity=CoreObject.removeEntity(self, entity)          # run the default method     
        self.waitToDispose=False                            # update the waitToDispose flag
        # if the Machine canAccept then signal a giver
        if self.canAccept():
            self.printTrace(self.id, attemptSignalGiver='(removeEntity)')
            self.signalGiver()
        return activeEntity
    
    # ======================================================================= 
    # checks if the Machine can dispose an entity to the following object
    # =======================================================================
    def haveToDispose(self, callerObject=None):
        activeObjectQueue=self.Res.users
        #if we have only one successor just check if machine waits to dispose and also is up
        # this is done to achieve better (cpu) processing time        
        if(callerObject==None):
            return len(activeObjectQueue)>0 and self.waitToDispose and (self.canDeliverOnInterruption or self.checkIfActive())
        thecaller=callerObject
        return len(activeObjectQueue)>0\
             and self.waitToDispose\
             and (thecaller in self.next)\
             and (self.canDeliverOnInterruption or self.checkIfActive())
                 
    # =======================================================================
    #                       calculates the setup time
    # =======================================================================
    def calculateSetupTime(self):
        return self.stpRng.generateNumber()
    
    # =======================================================================
    #                        calculates the Load time
    # =======================================================================
    def calculateLoadTime(self):
        return self.loadRng.generateNumber()
    
            
    #===========================================================================
    # find candidate operators within the free operators
    #===========================================================================
    def findCandidateOperator(self):
        # TODO: this way no sorting is performed
        # find an available operator
        candidateOperator=self.operatorPool.findAvailableOperator()
        # append the station into its candidateStations
        candidateOperator.candidateStations.append(self)
        return candidateOperator
    
    #===========================================================================
    # checks whether the entity can proceed to a successor object
    #===========================================================================
    def canDeliver(self, entity=None):
        assert self.isInActiveQueue(entity), entity.id +' not in the internalQueue of'+ self.id
        activeEntity=entity
        
        from Globals import G
        router = G.Router
        # if the entity is in a machines who's broker waits for operator then
        if self in router.pendingMachines:
            activeEntity.proceed=True
            activeEntity.candidateReceivers.append(self)
            return True
        return False
    
    # =======================================================================
    #                   prepare the machine to be operated
    # =======================================================================
    def requestOperator(self):
        self.broker.invoke()
        self.toBeOperated = True
    
    # =======================================================================
    #                   prepare the machine to be released
    # =======================================================================
    def releaseOperator(self):
        self.outputTrace(self.currentOperator.objName, "released from "+ self.objName)
        # set the flag operatorAssignedTo to None
        # XXX in case of skilled operators which stay at the same station should that change
        if not self.checkForDedicatedOperators():
            self.currentOperator.unAssign()            
        self.broker.invoke()
        self.toBeOperated = False
        
    # =======================================================================
    #       check if the machine is currently operated by an operator
    # =======================================================================
    def isOperated(self):
        return self.toBeOperated
    
    # =======================================================================
    # actions to be taken after the simulation ends
    # =======================================================================
    def postProcessing(self, MaxSimtime=None): 
        # ToDo code of the extension does not currently do anything. Check how we want to count waitingForOperator etc time
        CoreObject.postProcessing(self, MaxSimtime)
        
        activeObject=self.getActiveObject()
        activeObjectQueue=self.getActiveObjectQueue()
               
        #calculate the offShift time for current entity
        offShiftTimeInCurrentEntity=0
        if self.interruptedBy:
            if self.onShift==False: # and self.interruptedBy=='ShiftScheduler':
                offShiftTimeInCurrentEntity=self.env.now-activeObject.timeLastShiftEnded

        
        elif(len(activeObject.getActiveObjectQueue())>0)\
            and (not (activeObject.nameLastEntityEnded==activeObject.nameLastEntityEntered))\
            and (activeObject.currentOperator==None):
            # TODO: needs further research as the time of failure while waiting for operator is not counted yet
            if self.Up==False:
                activeObject.downTimeProcessingCurrentEntity+=self.env.now-activeObject.timeLastFailure
            activeObject.totalTimeWaitingForOperator+=self.env.now-activeObject.timeWaitForOperatorStarted\
                                                           -activeObject.downTimeProcessingCurrentEntity\
                                                           -offShiftTimeInCurrentEntity
     
    # =======================================================================
    # outputs the the "output.xls"
    # =======================================================================
    def outputResultsXL(self, MaxSimtime=None):
        from Globals import G
        from Globals import getConfidenceIntervals
        if MaxSimtime==None:
            MaxSimtime=G.maxSimTime
        
        if(G.numberOfReplications==1): #if we had just one replication output the results to excel    
            G.outputSheet.write(G.outputIndex,0, "The percentage of Failure of " +self.objName+ " is:")
            G.outputSheet.write(G.outputIndex,1,100*self.totalFailureTime/MaxSimtime)
            G.outputIndex+=1 
            G.outputSheet.write(G.outputIndex,0, "The percentage of Working of " +self.objName+ " is:")
            G.outputSheet.write(G.outputIndex,1,100*self.totalWorkingTime/MaxSimtime)
            G.outputIndex+=1 
            G.outputSheet.write(G.outputIndex,0, "The percentage of Blockage of " +self.objName+ " is:")
            G.outputSheet.write(G.outputIndex,1,100*self.totalBlockageTime/MaxSimtime)
            G.outputIndex+=1 
            G.outputSheet.write(G.outputIndex,0, "The percentage of Waiting of " +self.objName+ " is:")
            G.outputSheet.write(G.outputIndex,1,100*self.totalWaitingTime/MaxSimtime)
            G.outputIndex+=1       
        else:        #if we had multiple replications we output confidence intervals to excel
                #for some outputs the results may be the same for each run (eg model is stochastic but failures fixed
                #so failurePortion will be exactly the same in each run). That will give 0 variability and errors.
                #so for each output value we check if there was difference in the runs' results
                #if yes we output the Confidence Intervals. if not we output just the fix value    
            
            G.outputSheet.write(G.outputIndex,0, "CI "+str(G.confidenceLevel*100)+"% for the mean percentage of Failure of "+ self.objName+" is:")
            failure_ci = getConfidenceIntervals(self.Failure)
            G.outputSheet.write(G.outputIndex, 1, failure_ci['min'])
            G.outputSheet.write(G.outputIndex, 2, failure_ci['avg'])
            G.outputSheet.write(G.outputIndex, 3, failure_ci['max'])
            G.outputIndex+=1

            G.outputSheet.write(G.outputIndex,0, "CI "+str(G.confidenceLevel*100)+"% for the mean percentage of Working of "+ self.objName+" is:")
            working_ci = getConfidenceIntervals(self.Working)
            G.outputSheet.write(G.outputIndex, 1, working_ci['min'])
            G.outputSheet.write(G.outputIndex, 2, working_ci['avg'])
            G.outputSheet.write(G.outputIndex, 3, working_ci['max'])
            G.outputIndex+=1

            G.outputSheet.write(G.outputIndex,0, "CI "+str(G.confidenceLevel*100)+"% for the mean percentage of Blockage of "+ self.objName+" is:")
            blockage_ci = getConfidenceIntervals(self.Blockage)
            G.outputSheet.write(G.outputIndex, 1, blockage_ci['min'])
            G.outputSheet.write(G.outputIndex, 2, blockage_ci['avg'])
            G.outputSheet.write(G.outputIndex, 3, blockage_ci['max'])
            G.outputIndex+=1

            G.outputSheet.write(G.outputIndex,0, "CI "+str(G.confidenceLevel*100)+"% for the mean percentage of Waiting of "+ self.objName+" is:")
            waiting_ci = getConfidenceIntervals(self.Waiting)
            G.outputSheet.write(G.outputIndex, 1, waiting_ci['min'])
            G.outputSheet.write(G.outputIndex, 2, waiting_ci['avg'])
            G.outputSheet.write(G.outputIndex, 3, waiting_ci['max'])
            G.outputIndex+=1
        G.outputIndex+=1
    
    # =======================================================================    
    # outputs results to JSON File
    # =======================================================================
    def outputResultsJSON(self):
        from Globals import G
        from Globals import getConfidenceIntervals
        json = {'_class': 'Dream.%s' % self.__class__.__name__,
                'id': self.id,
                'family': self.family,
                'results': {}}
        if (G.numberOfReplications == 1):
            # if we had just one replication output the results as numbers
            json['results']['failure_ratio']=100*self.totalFailureTime/G.maxSimTime
            json['results']['working_ratio']=100*self.totalWorkingTime/G.maxSimTime
            json['results']['blockage_ratio']=100*self.totalBlockageTime/G.maxSimTime
            json['results']['waiting_ratio']=100*self.totalWaitingTime/G.maxSimTime
            #output the off-shift time only if there is any
            if self.totalOffShiftTime:
                json['results']['off_shift_ratio']=100*self.totalOffShiftTime/G.maxSimTime
            if any(type=='Setup' for type in self.multOperationTypeList):
                json['results']['setup_ratio']=100*self.totalSetupTime/G.maxSimTime
            if any(type=='Load' for type in self.multOperationTypeList):
                json['results']['load_ratio']=100*self.totalLoadTime/G.maxSimTime
        else:
            json['results']['failure_ratio'] = getConfidenceIntervals(self.Failure)
            json['results']['working_ratio'] = getConfidenceIntervals(self.Working)
            json['results']['blockage_ratio'] = getConfidenceIntervals(self.Blockage)
            json['results']['waiting_ratio'] = getConfidenceIntervals(self.Waiting)
            json['results']['off_shift_ratio'] = getConfidenceIntervals(self.OffShift)
            json['results']['setup_ratio'] = getConfidenceIntervals(self.SettingUp)
            json['results']['loading_ratio'] = getConfidenceIntervals(self.Loading)

        G.outputJSON['elementList'].append(json)

